# Converted from <asm-generic/ioctl.h>

# ioctl command encoding: 32 bits total, command in lower 16 bits,
# size of the parameter structure in the lower 14 bits of the
# upper 16 bits.
#
# Encoding the size of the parameter structure in the ioctl request
# is useful for catching programs compiled with old versions
# and to avoid overwriting user space outside the user buffer area.
# The highest 2 bits are reserved for indicating the ``access mode''.
#
# NOTE: This limits the max parameter size to 16kB -1 !

# The following is for compatibility across the various Linux
# platforms.  The generic ioctl numbering scheme doesn't really enforce
# a type field.  De facto, however, the top 8 bits of the lower 16
# bits are indeed used as a type field, so we might just as well make
# this explicit here.  Please be sure to use the decoding macros
# below from now on.

import ctypes

_IOC_NRBITS =	8
_IOC_TYPEBITS =	8

_IOC_SIZEBITS =	14
_IOC_DIRBITS =	2

_IOC_NRMASK =   (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK =  (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT =	 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT =	 _IOC_SIZESHIFT + _IOC_SIZEBITS

# Direction bits

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2

def _IOC(dir, type, nr, size):
    return (dir  << _IOC_DIRSHIFT) | \
           (type << _IOC_TYPESHIFT) | \
           (nr   << _IOC_NRSHIFT) | \
           (size << _IOC_SIZESHIFT)

def _IOC_TYPECHECK(t):
    return ctypes.sizeof(t)


# used to create ioctl numbers

def _IO(type, nr):
    return _IOC(_IOC_NONE, type, nr, 0)

def _IOR(type, nr, size):
    return _IOC(_IOC_READ, type, nr, _IOC_TYPECHECK(size))

def _IOW(type, nr, size):
    return _IOC(_IOC_WRITE, type, nr, _IOC_TYPECHECK(size))

def _IOWR(type,nr,size):
    return _IOC(_IOC_READ|_IOC_WRITE, type, nr, _IOC_TYPECHECK(size))

def _IOR_BAD(type,nr,size):
    return _IOC(_IOC_READ, type, nr, sizeof(size))

def _IOW_BAD(type,nr,size):
    return _IOC(_IOC_WRITE,type,nr, sizeof(size))

def _IOWR_BAD(type,nr,size):
    return _IOC(_IOC_READ|_IOC_WRITE, type, nr, sizeof(size))


# ...and for the drivers/sound files...

IOC_IN = _IOC_WRITE << _IOC_DIRSHIFT
IOC_OUT = _IOC_READ << _IOC_DIRSHIFT
IOC_INOUT = (_IOC_WRITE|_IOC_READ) << _IOC_DIRSHIFT
IOCSIZE_MASK = _IOC_SIZEMASK << _IOC_SIZESHIFT
IOCSIZE_SHIFT = _IOC_SIZESHIFT




import posix
import ctypes
from fcntl import ioctl
from .linux_spi_spidev import spi_ioc_transfer, SPI_IOC_MESSAGE


SPIDEV = '/dev/spidev'
SPI_HELP_LINK = "http://piface.github.io/pifacecommon/installation.html" \
    "#enable-the-spi-module"


class SPIInitError(Exception):
    pass


class SPIDevice(object):
    """An SPI Device at /dev/spi<bus>.<chip_select>."""
    def __init__(self, bus=0, chip_select=0, spi_callback=None):
        """Initialises the SPI device file descriptor.

        :param bus: The SPI device bus number
        :type bus: int
        :param chip_select: The SPI device chip_select number
        :param chip_select: int
        :raises: InitError
        """
        self.bus = bus
        self.chip_select = chip_select
        self.spi_callback = spi_callback
        self.fd = None
        spi_device = "%s%d.%d" % (SPIDEV, self.bus, self.chip_select)
        self.open_fd(spi_device)

    # def __del__(self):
    #     if self.fd is not None:
    #         self.close_fd()

    def open_fd(self, spi_device):
        try:
            self.fd = posix.open(spi_device, posix.O_RDWR)
        except OSError as e:
            raise SPIInitError(
                "I can't see %s. Have you enabled the SPI module? (%s)"
                % (spi_device, SPI_HELP_LINK)
            )  # from e  # from is only available in Python 3

    def close_fd(self):
        posix.close(self.fd)
        self.fd = None

    def spisend(self, bytes_to_send):
        """Sends bytes via the SPI bus.

        :param bytes_to_send: The bytes to send on the SPI device.
        :type bytes_to_send: bytes
        :returns: bytes -- returned bytes from SPI device
        :raises: InitError
        """
        # make some buffer space to store reading/writing
        wbuffer = ctypes.create_string_buffer(bytes_to_send,
                                              len(bytes_to_send))
        rbuffer = ctypes.create_string_buffer(len(bytes_to_send))

        # create the spi transfer struct
        transfer = spi_ioc_transfer(
            tx_buf=ctypes.addressof(wbuffer),
            rx_buf=ctypes.addressof(rbuffer),
            len=ctypes.sizeof(wbuffer)
        )

        if self.spi_callback is not None:
            self.spi_callback(bytes_to_send)
        # send the spi command
        ioctl(self.fd, SPI_IOC_MESSAGE(1), transfer)
        return ctypes.string_at(rbuffer, ctypes.sizeof(rbuffer))



# Converted from <linux/spi/spidev.h>
import ctypes
from .asm_generic_ioctl import _IOR, _IOW, _IOC_SIZEBITS

SPI_CPHA = 0x01
SPI_CPOL = 0x02

SPI_MODE_0 = 0
SPI_MODE_1 = SPI_CPHA
SPI_MODE_2 = SPI_CPOL
SPI_MODE_3 = SPI_CPOL | SPI_CPHA

SPI_CS_HIGH = 0x04
SPI_LSB_FIRST = 0x08
SPI_3WIRE = 0x10
SPI_LOOP = 0x20
SPI_NO_CS = 0x40
SPI_READY = 0x80


# IOCTL commands */

SPI_IOC_MAGIC = 107  # ord('k')


# struct spi_ioc_transfer - describes a single SPI transfer
#
# tx_buf:        Holds pointer to userspace buffer with transmit data, or null.
#                If no data is provided, zeroes are shifted out.
# rx_buf:        Holds pointer to userspace buffer for receive data, or null.
# len:           Length of tx and rx buffers, in bytes.
# speed_hz:      Temporary override of the device's bitrate.
# bits_per_word: Temporary override of the device's wordsize.
# delay_usecs:   If nonzero, how long to delay after the last bit transfer
#            before optionally deselecting the device before the next transfer.
# cs_change:     True to deselect device before starting the next transfer.
#
# This structure is mapped directly to the kernel spi_transfer structure;
# the fields have the same meanings, except of course that the pointers
# are in a different address space (and may be of different sizes in some
# cases, such as 32-bit i386 userspace over a 64-bit x86_64 kernel).
# Zero-initialize the structure, including currently unused fields, to
# accomodate potential future updates.
#
# SPI_IOC_MESSAGE gives userspace the equivalent of kernel spi_sync().
# Pass it an array of related transfers, they'll execute together.
# Each transfer may be half duplex (either direction) or full duplex.
#
#   struct spi_ioc_transfer mesg[4];
#   ...
#   status = ioctl(fd, SPI_IOC_MESSAGE(4), mesg);
#
# So for example one transfer might send a nine bit command (right aligned
# in a 16-bit word), the next could read a block of 8-bit data before
# terminating that command by temporarily deselecting the chip; the next
# could send a different nine bit command (re-selecting the chip), and the
# last transfer might write some register values.

class spi_ioc_transfer(ctypes.Structure):
    """<linux/spi/spidev.h> struct spi_ioc_transfer"""

    _fields_ = [
        ("tx_buf", ctypes.c_uint64),
        ("rx_buf", ctypes.c_uint64),
        ("len", ctypes.c_uint32),
        ("speed_hz", ctypes.c_uint32),
        ("delay_usecs", ctypes.c_uint16),
        ("bits_per_word", ctypes.c_uint8),
        ("cs_change", ctypes.c_uint8),
        ("pad", ctypes.c_uint32)]

    __slots__ = [name for name, type in _fields_]


# not all platforms use <asm-generic/ioctl.h> or _IOC_TYPECHECK() ...
def SPI_MSGSIZE(N):
    if ((N)*(ctypes.sizeof(spi_ioc_transfer))) < (1 << _IOC_SIZEBITS):
        return (N)*(ctypes.sizeof(spi_ioc_transfer))
    else:
        return 0


def SPI_IOC_MESSAGE(N):
    return _IOW(SPI_IOC_MAGIC, 0, ctypes.c_char*SPI_MSGSIZE(N))


# Read / Write of SPI mode (SPI_MODE_0..SPI_MODE_3)
SPI_IOC_RD_MODE = _IOR(SPI_IOC_MAGIC, 1, ctypes.c_uint8)
SPI_IOC_WR_MODE = _IOW(SPI_IOC_MAGIC, 1, ctypes.c_uint8)

# Read / Write SPI bit justification
SPI_IOC_RD_LSB_FIRST = _IOR(SPI_IOC_MAGIC, 2, ctypes.c_uint8)
SPI_IOC_WR_LSB_FIRST = _IOW(SPI_IOC_MAGIC, 2, ctypes.c_uint8)

# Read / Write SPI device word length (1..N)
SPI_IOC_RD_BITS_PER_WORD = _IOR(SPI_IOC_MAGIC, 3, ctypes.c_uint8)
SPI_IOC_WR_BITS_PER_WORD = _IOW(SPI_IOC_MAGIC, 3, ctypes.c_uint8)

# Read / Write SPI device default max speed hz
SPI_IOC_RD_MAX_SPEED_HZ = _IOR(SPI_IOC_MAGIC, 4, ctypes.c_uint32)
SPI_IOC_WR_MAX_SPEED_HZ = _IOW(SPI_IOC_MAGIC, 4, ctypes.c_uint32)



def get_bit_mask(bit_num):
    """Returns as bit mask with bit_num set.

    :param bit_num: The bit number.
    :type bit_num: int
    :returns: int -- the bit mask
    :raises: RangeError

    >>> bin(pifacecommon.core.get_bit_mask(0))
    1
    >>> pifacecommon.core.get_bit_mask(1)
    2
    >>> bin(pifacecommon.core.get_bit_mask(3))
    '0b1000'
    """
    return 1 << (bit_num)


def get_bit_num(bit_pattern):
    """Returns the lowest bit num from a given bit pattern. Returns None if no
    bits set.

    :param bit_pattern: The bit pattern.
    :type bit_pattern: int
    :returns: int -- the bit number
    :returns: None -- no bits set

    >>> pifacecommon.core.get_bit_num(0)
    None
    >>> pifacecommon.core.get_bit_num(0b1)
    0
    >>> pifacecommon.core.get_bit_num(0b11000)
    3
    """
    if bit_pattern == 0:
        return None

    bit_num = 0  # assume bit 0
    while (bit_pattern & 1) == 0:
        bit_pattern = bit_pattern >> 1
        bit_num += 1
        if bit_num > 7:
            bit_num = 0
            break

    return bit_num


def sleep_microseconds(microseconds):
    """Sleeps for the given number of microseconds.

    :param microseconds: Number of microseconds to sleep for.
    :type microseconds: int
    """
    # divide microseconds by 1 million for seconds
    seconds = microseconds / float(1000000)
    time.sleep(seconds)



import os
import sys
from .core import (
    get_bit_mask,
    get_bit_num,
)
from .spi import SPIDevice
import pifacecommon.interrupts


# Python 2 support
PY3 = sys.version_info[0] >= 3

WRITE_CMD = 0
READ_CMD = 1

# Register addresses
IODIRA = 0x0  # I/O direction A
IODIRB = 0x1  # I/O direction B
IPOLA = 0x2  # I/O polarity A
IPOLB = 0x3  # I/O polarity B
GPINTENA = 0x4  # interupt enable A
GPINTENB = 0x5  # interupt enable B
DEFVALA = 0x6  # register default value A (interupts)
DEFVALB = 0x7  # register default value B (interupts)
INTCONA = 0x8  # interupt control A
INTCONB = 0x9  # interupt control B
IOCON = 0xA  # I/O config (also 0xB)
GPPUA = 0xC  # port A pullups
GPPUB = 0xD  # port B pullups
INTFA = 0xE  # interupt flag A (where the interupt came from)
INTFB = 0xF  # interupt flag B
INTCAPA = 0x10  # interupt capture A (value at interupt is saved here)
INTCAPB = 0x11  # interupt capture B
GPIOA = 0x12  # port A
GPIOB = 0x13  # port B
OLATA = 0x14  # output latch A
OLATB = 0x15  # output latch B

# I/O config
BANK_OFF = 0x00  # addressing mode
BANK_ON = 0x80
INT_MIRROR_ON = 0x40  # interupt mirror (INTa|INTb)
INT_MIRROR_OFF = 0x00
SEQOP_OFF = 0x20  # incrementing address pointer
SEQOP_ON = 0x00
DISSLW_ON = 0x10  # slew rate
DISSLW_OFF = 0x00
HAEN_ON = 0x08  # hardware addressing
HAEN_OFF = 0x00
ODR_ON = 0x04  # open drain for interupts
ODR_OFF = 0x00
INTPOL_HIGH = 0x02  # interupt polarity
INTPOL_LOW = 0x00

LOWER_NIBBLE, UPPER_NIBBLE = range(2)


class MCP23S17(SPIDevice):
    """Microchip's MCP23S17: A 16-Bit I/O Expander with Serial Interface.

    :attribute: iodira/iodirb -- Controls the direction of the data I/O.
    :attribute: ipola/ipolb --This register allows the user to configure
                              the polarity on the corresponding GPIO port
                              bits.
    :attribute: gpintena/gpintenb -- The GPINTEN register controls the
                                     interrupt-onchange feature for each
                                     pin.
    :attribute: defvala/defvalb --The default comparison value is
                                  configured in the DEFVAL register.
    :attribute: intcona/intconb --The INTCON register controls how the
                                  associated pin value is compared for
                                  the interrupt-on-change feature.
    :attribute: iocon --The IOCON register contains several bits for
                        configuring the device.
    :attribute: gppua/gppub --The GPPU register controls the pull-up
                              resistors for the port pins.
    :attribute: intfa/intfb --The INTF register reflects the interrupt
                              condition on the port pins of any pin that
                              is enabled for interrupts via the GPINTEN
                              register.
    :attribute: intcapa/intcapb -- The INTCAP register captures the GPIO
                                   port value at the time the interrupt
                                   occurred.
    :attribute: gpioa/gpiob -- The GPIO register reflects the value on
                               the port.
    :attribute: olata/olatb -- The OLAT register provides access to the
                               output latches.
    """
    def __init__(self, hardware_addr=0, bus=0, chip_select=0):
        super(MCP23S17, self).__init__(bus, chip_select)
        self.hardware_addr = hardware_addr

        self.iodira = MCP23S17Register(IODIRA, self)
        self.iodirb = MCP23S17Register(IODIRB, self)
        self.ipola = MCP23S17Register(IPOLA, self)
        self.ipolb = MCP23S17Register(IPOLB, self)
        self.gpintena = MCP23S17Register(GPINTENA, self)
        self.gpintenb = MCP23S17Register(GPINTENB, self)
        self.defvala = MCP23S17Register(DEFVALA, self)
        self.defvalb = MCP23S17Register(DEFVALB, self)
        self.intcona = MCP23S17Register(INTCONA, self)
        self.intconb = MCP23S17Register(INTCONB, self)
        self.iocon = MCP23S17Register(IOCON, self)
        self.gppua = MCP23S17Register(GPPUA, self)
        self.gppub = MCP23S17Register(GPPUB, self)
        self.intfa = MCP23S17Register(INTFA, self)
        self.intfb = MCP23S17Register(INTFB, self)
        self.intcapa = MCP23S17Register(INTCAPA, self)
        self.intcapb = MCP23S17Register(INTCAPB, self)
        self.gpioa = MCP23S17Register(GPIOA, self)
        self.gpiob = MCP23S17Register(GPIOB, self)
        self.olata = MCP23S17Register(OLATA, self)
        self.olatb = MCP23S17Register(OLATB, self)

    def _get_spi_control_byte(self, read_write_cmd):
        """Returns an SPI control byte.

        The MCP23S17 is a slave SPI device. The slave address contains
        four fixed bits and three user-defined hardware address bits
        (if enabled via IOCON.HAEN) (pins A2, A1 and A0) with the
        read/write bit filling out the control byte::

            +--------------------+
            |0|1|0|0|A2|A1|A0|R/W|
            +--------------------+
             7 6 5 4 3  2  1   0

        :param read_write_cmd: Read or write command.
        :type read_write_cmd: int
        """
        # board_addr_pattern = (self.hardware_addr & 0b111) << 1
        board_addr_pattern = (self.hardware_addr << 1) & 0xE
        rw_cmd_pattern = read_write_cmd & 1  # make sure it's just 1 bit long
        return 0x40 | board_addr_pattern | rw_cmd_pattern

    def read(self, address):
        """Returns the value of the address specified.

        :param address: The address to read from.
        :type address: int
        """
        return self._pyver_read(address)

    def _py3read(self, address):
        ctrl_byte = self._get_spi_control_byte(READ_CMD)
        op, addr, data = self.spisend(bytes((ctrl_byte, address, 0)))
        return data

    def _py2read(self, address):
        ctrl_byte = self._get_spi_control_byte(READ_CMD)
        op, addr, data = self.spisend(chr(ctrl_byte)+chr(address)+chr(0))
        return ord(data)

    def write(self, data, address):
        """Writes data to the address specified.

        :param data: The data to write.
        :type data: int
        :param address: The address to write to.
        :type address: int
        """
        self._pyver_write(data, address)

    def _py3write(self, data, address):
        ctrl_byte = self._get_spi_control_byte(WRITE_CMD)
        self.spisend(bytes((ctrl_byte, address, data)))

    def _py2write(self, data, address):
        ctrl_byte = self._get_spi_control_byte(WRITE_CMD)
        self.spisend(chr(ctrl_byte)+chr(address)+chr(data))

    # Python 2 support
    _pyver_read = _py3read if PY3 else _py2read
    _pyver_write = _py3write if PY3 else _py2write

    def read_bit(self, bit_num, address):
        """Returns the bit specified from the address.

        :param bit_num: The bit number to read from.
        :type bit_num: int
        :param address: The address to read from.
        :type address: int
        :returns: int -- the bit value from the address
        """
        value = self.read(address)
        bit_mask = get_bit_mask(bit_num)
        return 1 if value & bit_mask else 0

    def write_bit(self, value, bit_num, address):
        """Writes the value given to the bit in the address specified.

        :param value: The value to write.
        :type value: int
        :param bit_num: The bit number to write to.
        :type bit_num: int
        :param address: The address to write to.
        :type address: int
        """
        bit_mask = get_bit_mask(bit_num)
        old_byte = self.read(address)
         # generate the new byte
        if value:
            new_byte = old_byte | bit_mask
        else:
            new_byte = old_byte & ~bit_mask
        self.write(new_byte, address)

    def clear_interrupts(self, port):
        """Clears the interrupt flags by 'read'ing the capture register."""
        self.read(INTCAPA if port == GPIOA else INTCAPB)


class MCP23S17RegisterBase(object):
    """Base class for objects on an 8-bit register inside an MCP23S17."""
    def __init__(self, address, chip):
        self.address = address
        self.chip = chip


class MCP23S17Register(MCP23S17RegisterBase):
    """An 8-bit register inside an MCP23S17."""
    def __init__(self, address, chip):
        super(MCP23S17Register, self).__init__(address, chip)
        self.lower_nibble = MCP23S17RegisterNibble(LOWER_NIBBLE, self.address,
                                                   self.chip)
        self.upper_nibble = MCP23S17RegisterNibble(UPPER_NIBBLE, self.address,
                                                   self.chip)
        self.bits = [MCP23S17RegisterBit(i, self.address, self.chip)
                     for i in range(8)]

    @property
    def value(self):
        return self.chip.read(self.address)

    @value.setter
    def value(self, v):
        self.chip.write(v, self.address)

    def all_high(self):
        self.value = 0xFF

    def all_low(self):
        self.value = 0x00

    all_on = all_high
    all_off = all_low

    def toggle(self):
        self.value = 0xFF ^ self.value


class MCP23S17RegisterNeg(MCP23S17Register):
    """An negated 8-bit register inside an MCP23S17."""
    def __init__(self, address, chip):
        super(MCP23S17RegisterNeg, self).__init__(address, chip)
        self.lower_nibble = MCP23S17RegisterNibbleNeg(LOWER_NIBBLE,
                                                      self.address,
                                                      self.chip)
        self.upper_nibble = MCP23S17RegisterNibbleNeg(UPPER_NIBBLE,
                                                      self.address,
                                                      self.chip)
        self.bits = [MCP23S17RegisterBitNeg(i, self.address, self.chip)
                     for i in range(8)]

    @property
    def value(self):
        return 0xFF ^ self.chip.read(self.address)

    @value.setter
    def value(self, v):
        self.chip.write(0xFF ^ v, self.address)


class MCP23S17RegisterNibble(MCP23S17RegisterBase):
    """An 4-bit nibble inside a register inside an MCP23S17."""
    def __init__(self, nibble, address, chip):
        super(MCP23S17RegisterNibble, self).__init__(address, chip)
        self.nibble = nibble
        range_start = 0 if self.nibble == LOWER_NIBBLE else 4
        range_end = 4 if self.nibble == LOWER_NIBBLE else 8
        self.bits = [MCP23S17RegisterBit(i, self.address, self.chip)
                     for i in range(range_start, range_end)]

    @property
    def value(self):
        if self.nibble == LOWER_NIBBLE:
            return self.chip.read(self.address) & 0xF
        elif self.nibble == UPPER_NIBBLE:
            return (self.chip.read(self.address) & 0xF0) >> 4

    @value.setter
    def value(self, v):
        register_value = self.chip.read(self.address)
        if self.nibble == LOWER_NIBBLE:
            register_value &= 0xF0  # clear
            register_value ^= (v & 0x0F)  # set
        elif self.nibble == UPPER_NIBBLE:
            register_value &= 0x0F  # clear
            register_value ^= ((v << 4) & 0xF0)  # set
        self.chip.write(register_value, self.address)

    def all_high(self):
        self.value = 0xF

    def all_low(self):
        self.value = 0x0

    all_on = all_high
    all_off = all_low

    def toggle(self):
        self.value = 0xF ^ self.value


class MCP23S17RegisterNibbleNeg(MCP23S17RegisterNibble):
    """A negated 4-bit nibble inside a register inside an MCP23S17."""
    def __init__(self, nibble, address, chip):
        super(MCP23S17RegisterNibbleNeg, self).__init__(nibble, address, chip)
        self.nibble = nibble
        range_start = 0 if self.nibble == LOWER_NIBBLE else 4
        range_end = 4 if self.nibble == LOWER_NIBBLE else 8
        self.bits = [MCP23S17RegisterBitNeg(i, self.address, self.chip)
                     for i in range(range_start, range_end)]

    @property
    def value(self):
        if self.nibble == LOWER_NIBBLE:
            v = self.chip.read(self.address) & 0xF
        elif self.nibble == UPPER_NIBBLE:
            v = (self.chip.read(self.address) & 0xF0) >> 4
        return 0xF ^ v

    @value.setter
    def value(self, v):
        register_value = self.chip.read(self.address)
        if self.nibble == LOWER_NIBBLE:
            register_value &= 0xF0  # clear
            register_value ^= (v & 0x0F ^ 0x0F)  # set
        elif self.nibble == UPPER_NIBBLE:
            register_value &= 0x0F  # clear
            register_value ^= ((v << 4) & 0xF0 ^ 0xF0)  # set
        self.chip.write(register_value, self.address)


class MCP23S17RegisterBit(MCP23S17RegisterBase):
    """A bit inside register inside an MCP23S17."""
    def __init__(self, bit_num, address, chip):
        super(MCP23S17RegisterBit, self).__init__(address, chip)
        self.bit_num = bit_num

    @property
    def value(self):
        return self.chip.read_bit(self.bit_num, self.address)

    @value.setter
    def value(self, v):
        self.chip.write_bit(v, self.bit_num, self.address)

    def set_high(self):
        self.value = 1

    def set_low(self):
        self.value = 0

    turn_on = set_high
    turn_off = set_low

    def toggle(self):
        self.value = 1 ^ self.value


class MCP23S17RegisterBitNeg(MCP23S17RegisterBit):
    """A negated bit inside register inside an MCP23S17."""
    def __init__(self, bit_num, address, chip):
        super(MCP23S17RegisterBitNeg, self).__init__(bit_num, address, chip)

    @property
    def value(self):
        return 1 ^ self.chip.read_bit(self.bit_num, self.address)

    @value.setter
    def value(self, v):
        self.chip.write_bit(v ^ 1, self.bit_num, self.address)



import threading
import multiprocessing
import select
import time
import errno
from .core import get_bit_num
import pifacecommon.mcp23s17


# interrupts
IODIR_FALLING_EDGE = IODIR_ON = 0
IODIR_RISING_EDGE = IODIR_OFF = 1
IODIR_BOTH = None
# IN_EVENT_DIR_OFF = INPUT_DIRECTION_OFF = 1
# IN_EVENT_DIR_BOTH = INPUT_DIRECTION_BOTH = None

GPIO_INTERRUPT_PIN = 25
GPIO_INTERRUPT_DEVICE = "/sys/class/gpio/gpio%d" % GPIO_INTERRUPT_PIN
GPIO_INTERRUPT_DEVICE_EDGE = '%s/edge' % GPIO_INTERRUPT_DEVICE
GPIO_INTERRUPT_DEVICE_VALUE = '%s/value' % GPIO_INTERRUPT_DEVICE
GPIO_EXPORT_FILE = "/sys/class/gpio/export"
GPIO_UNEXPORT_FILE = "/sys/class/gpio/unexport"

# max seconds to wait for file I/O (when enabling interrupts)
FILE_IO_TIMEOUT = 1

READY_FOR_EVENTS = "ready for events"

# deboucing
DEFAULT_SETTLE_TIME = 0.020  # 20ms


class Timeout(Exception):
    pass


class InterruptEvent(object):
    """An interrupt event containting the interrupt flag and capture register
    values, the chip object from which the interrupt occured and a timestamp.
    """
    def __init__(
            self, interrupt_flag, interrupt_capture, chip, timestamp):
        self.interrupt_flag = interrupt_flag
        self.interrupt_capture = interrupt_capture
        self.chip = chip
        self.timestamp = timestamp

    def __str__(self):
        s = "interrupt_flag:    {flag}\n" \
            "interrupt_capture: {capture}\n" \
            "pin_num:           {pin_num}\n" \
            "direction:         {direction}\n" \
            "chip:              {chip}\n" \
            "timestamp:         {timestamp}"
        return s.format(flag=bin(self.interrupt_flag),
                        capture=bin(self.interrupt_capture),
                        pin_num=self.pin_num,
                        direction=self.direction,
                        chip=self.chip,
                        timestamp=self.timestamp)

    @property
    def pin_num(self):
        return get_bit_num(self.interrupt_flag)

    @property
    def direction(self):
        return (self.interrupt_flag & self.interrupt_capture) >> self.pin_num


class FunctionMap(object):
    """Maps something to a callback function.
    (This is an abstract class, you must implement a SomethingFunctionMap).
    """
    def __init__(self, callback, settle_time=None):
        self.callback = callback
        self.settle_time = settle_time


class PinFunctionMap(FunctionMap):
    """Maps an IO pin and a direction to callback function."""
    def __init__(self, pin_num, direction, callback, settle_time):
        self.pin_num = pin_num
        self.direction = direction
        super(PinFunctionMap, self).__init__(callback, settle_time)

    def __str__(self):
        s = "Pin num:  {pin_num}\n" \
            "Direction:{direction}\n"
        return s.format(pin_num=self.pin_num,
                        direction=self.direction)


class EventQueue(object):
    """Stores events in a queue."""
    def __init__(self, pin_function_maps):
        super(EventQueue, self).__init__()
        self.last_event_time = [0]*8  # last event time on each pin
        self.pin_function_maps = pin_function_maps
        self.queue = multiprocessing.Queue()

    def add_event(self, event):
        """Adds events to the queue. Will ignore events that occur before the
        settle time for that pin/direction. Such events are assumed to be
        bouncing.
        """
        # print("Trying to add event:")
        # print(event)
        # find out the pin settle time
        for pin_function_map in self.pin_function_maps:
            if _event_matches_pin_function_map(event, pin_function_map):
            # if pin_function_map.pin_num == event.pin_num and (
            #         pin_function_map.direction == event.direction or
            #         pin_function_map.direction == IODIR_BOTH):
                pin_settle_time = pin_function_map.settle_time
                # print("EventQueue: Found event in map.")
                break
        else:
            # Couldn't find event in map, don't bother adding it to the queue
            # print("EventQueue: Couldn't find event in map:")
            # for pin_function_map in self.pin_function_maps:
            #     print(pin_function_map)
            return

        threshold_time = self.last_event_time[event.pin_num] + pin_settle_time
        if event.timestamp > threshold_time:
            self.put(event)
            self.last_event_time[event.pin_num] = event.timestamp

    def put(self, thing):
        self.queue.put(thing)

    def get(self):
        return self.queue.get()


class PortEventListener(object):
    """Listens for port events and calls the registered functions.

    >>> def print_flag(event):
    ...     print(event.interrupt_flag)
    ...
    >>> port = pifacecommon.mcp23s17.GPIOA
    >>> listener = pifacecommon.interrupts.PortEventListener(port)
    >>> listener.register(0, pifacecommon.interrupts.IODIR_ON, print_flag)
    >>> listener.activate()
    """

    TERMINATE_SIGNAL = "astalavista"

    def __init__(self, port, chip, return_after_kbdint=True):
        self.port = port
        self.chip = chip
        self.pin_function_maps = list()
        self.event_queue = EventQueue(self.pin_function_maps)
        self.detector = multiprocessing.Process(
            target=watch_port_events,
            args=(
                self.port,
                self.chip,
                self.pin_function_maps,
                self.event_queue,
                return_after_kbdint))
        self.dispatcher = threading.Thread(
            target=handle_events,
            args=(
                self.pin_function_maps,
                self.event_queue,
                _event_matches_pin_function_map,
                PortEventListener.TERMINATE_SIGNAL))

    def register(self, pin_num, direction, callback,
                 settle_time=DEFAULT_SETTLE_TIME):
        """Registers a pin number and direction to a callback function.

        :param pin_num: The pin pin number.
        :type pin_num: int
        :param direction: The event direction
            (use: IODIR_ON/IODIR_OFF/IODIR_BOTH)
        :type direction: int
        :param callback: The function to run when event is detected.
        :type callback: function
        :param settle_time: Time within which subsequent events are ignored.
        :type settle_time: int
        """
        self.pin_function_maps.append(
            PinFunctionMap(pin_num, direction, callback, settle_time))

    def deregister(self, pin_num=None, direction=None):
        """De-registers callback functions

        :param pin_num: The pin number. If None then all functions are de-registered
        :type pin_num: int
        :param direction: The event direction. If None then all functions for the
                          given pin are de-registered
        :type direction:int
        """
        to_delete = []
        for i, function_map in enumerate(self.pin_function_maps):
            if ( pin_num == None
                 or ( function_map.pin_num == pin_num
                      and ( direction == None
                            or function_map.direction == direction ) ) ):
                to_delete.append(i)
        for i in reversed(to_delete):
            del self.pin_function_maps[i]

    def activate(self):
        """When activated the :class:`PortEventListener` will run callbacks
        associated with pins/directions.
        """
        self.detector.start()
        self.dispatcher.start()

    def deactivate(self):
        """When deactivated the :class:`PortEventListener` will not run
        anything.
        """
        self.event_queue.put(self.TERMINATE_SIGNAL)
        self.dispatcher.join()
        self.detector.terminate()
        self.detector.join()


class GPIOInterruptDevice(object):
    """A device that interrupts using the GPIO pins."""
    def gpio_interrupts_enable(self):
        """Enables GPIO interrupts."""
        try:
            bring_gpio_interrupt_into_userspace()
            set_gpio_interrupt_edge()
        except Timeout as e:
            raise InterruptEnableException(
                "There was an error bringing gpio%d into userspace. %s" %
                (GPIO_INTERRUPT_PIN, e.message)
            )

    def gpio_interrupts_disable(self):
        """Disables gpio interrupts."""
        set_gpio_interrupt_edge('none')
        deactivate_gpio_interrupt()


def _event_matches_pin_function_map(event, pin_function_map):
    # print("pin num", event.pin_num, pin_function_map.pin_num)
    # print("direction", event.direction, pin_function_map.direction)
    pin_match = event.pin_num == pin_function_map.pin_num
    direction_match = pin_function_map.direction is None
    direction_match |= event.direction == pin_function_map.direction
    return pin_match and direction_match


def watch_port_events(port, chip, pin_function_maps, event_queue,
                      return_after_kbdint=False):
    """Waits for a port event. When a port event occurs it is placed onto the
    event queue.

    :param port: The port we are waiting for interrupts on (GPIOA/GPIOB).
    :type port: int
    :param chip: The chip we are waiting for interrupts on.
    :type chip: :class:`pifacecommon.mcp23s17.MCP23S17`
    :param pin_function_maps: A list of classes that have inheritted from
        :class:`FunctionMap`\ s describing what to do with events.
    :type pin_function_maps: list
    :param event_queue: A queue to put events on.
    :type event_queue: :py:class:`multiprocessing.Queue`
    """
    # set up epoll
    gpio25 = open(GPIO_INTERRUPT_DEVICE_VALUE, 'r')  # change to use 'with'?
    epoll = select.epoll()
    epoll.register(gpio25, select.EPOLLIN | select.EPOLLET)

    while True:
        # wait here until input
        try:
            events = epoll.poll()
        except KeyboardInterrupt as e:
            if return_after_kbdint:
                return
            else:
                raise e
        except IOError as e:
            # ignore "Interrupted system call" error.
            # I don't really like this solution. Ignoring problems is bad!
            if e.errno != errno.EINTR:
                raise

        # find out where the interrupt came from and put it on the event queue
        if port == pifacecommon.mcp23s17.GPIOA:
            interrupt_flag = chip.intfa.value
        else:
            interrupt_flag = chip.intfb.value

        if interrupt_flag == 0:
            continue  # The interrupt has not been flagged on this board
        else:
            if port == pifacecommon.mcp23s17.GPIOA:
                interrupt_capture = chip.intcapa.value
            else:
                interrupt_capture = chip.intcapb.value
            event_queue.add_event(InterruptEvent(
                interrupt_flag, interrupt_capture, chip, time.time()))

    epoll.close()


def handle_events(
        function_maps, event_queue, event_matches_function_map,
        terminate_signal):
    """Waits for events on the event queue and calls the registered functions.

    :param function_maps: A list of classes that have inheritted from
        :class:`FunctionMap`\ s describing what to do with events.
    :type function_maps: list
    :param event_queue: A queue to put events on.
    :type event_queue: :py:class:`multiprocessing.Queue`
    :param event_matches_function_map: A function that determines if the given
        event and :class:`FunctionMap` match.
    :type event_matches_function_map: function
    :param terminate_signal: The signal that, when placed on the event queue,
        causes this function to exit.
    """
    while True:
        # print("HANDLE: Waiting for events!")
        event = event_queue.get()
        # print("HANDLE: It's an event!")
        if event == terminate_signal:
            return
        # if matching get the callback function, else function is None
        functions = map(
            lambda fm: fm.callback
            if event_matches_function_map(event, fm) else None,
            function_maps)
        # reduce to just the callback functions (remove None)
        # TODO: I think this can just be filter(None, functions)
        functions = filter(lambda f: f is not None, functions)

        for function in functions:
            function(event)


# def clear_interrupts(port):
#     """Clears the interrupt flags by 'read'ing the capture register
#     on all boards.
#     """
#     intcap = INTCAPA if port == GPIOA else INTCAPB
#     for i in range(MAX_BOARDS):
#         read(intcap, i)


# def enable_interrupts(port, pin_map=0xff, hardware_addrs=range(MAX_BOARDS)):
#     """Enables interrupts on the port specified. A pin map can be given to
#     only enable interrupts on some pins. A list of board numbers can also be
#     given to only enable the interrupts on some boards.

#     :param port: The port to enable interrupts on
#         (pifacecommon.core.GPIOA, pifacecommon.core.GPIOB)
#     :type port: int
#     :param pin_map: The pins to enable interrupts on
#     :type pin_map: int
#     :param hardware_addrs: The boards to enable interrupts on.
#     :type hardware_addrs: list
#     """
#     # enable interrupts
#     int_enable_port = GPINTENA if port == GPIOA else GPINTENB
#     for board_index in hardware_addrs:
#         write(pin_map, int_enable_port, board_index)

#     try:
#         bring_gpio_interrupt_into_userspace()
#         set_gpio_interrupt_edge()
#     except Timeout as e:
#         raise InterruptEnableException(
#             "There was an error bringing gpio%d into userspace. %s" %
#             (GPIO_INTERRUPT_PIN, e.message)
#         )


def bring_gpio_interrupt_into_userspace():  # activate gpio interrupt
    """Bring the interrupt pin on the GPIO into Linux userspace."""
    try:
        # is it already there?
        with open(GPIO_INTERRUPT_DEVICE_VALUE):
            return
    except IOError:
        # no, bring it into userspace
        with open(GPIO_EXPORT_FILE, 'w') as export_file:
            export_file.write(str(GPIO_INTERRUPT_PIN))

        wait_until_file_exists(GPIO_INTERRUPT_DEVICE_VALUE)


def deactivate_gpio_interrupt():
    """Remove the GPIO interrupt pin from Linux userspace."""
    with open(GPIO_UNEXPORT_FILE, 'w') as unexport_file:
        unexport_file.write(str(GPIO_INTERRUPT_PIN))


def set_gpio_interrupt_edge(edge='falling'):
    """Set the interrupt edge on the userspace GPIO pin.

    :param edge: The interrupt edge ('none', 'falling', 'rising').
    :type edge: string
    """
    # we're only interested in the falling edge (1 -> 0)
    start_time = time.time()
    time_limit = start_time + FILE_IO_TIMEOUT
    while time.time() < time_limit:
        try:
            with open(GPIO_INTERRUPT_DEVICE_EDGE, 'w') as gpio_edge:
                gpio_edge.write(edge)
                return
        except IOError:
            pass


def wait_until_file_exists(filename):
    """Wait until a file exists.

    :param filename: The name of the file to wait for.
    :type filename: string
    """
    start_time = time.time()
    time_limit = start_time + FILE_IO_TIMEOUT
    while time.time() < time_limit:
        try:
            with open(filename):
                return
        except IOError:
            pass

    raise Timeout("Waiting too long for %s." % filename)


# def disable_interrupts(port):
#     """Disables interrupts for all pins on the port specified.

#     :param port: The port to enable interrupts on
#         (pifacecommon.core.GPIOA, pifacecommon.core.GPIOB)
#     :type port: int
#     """
#     # neither edgez
#     with open(GPIO_INTERRUPT_DEVICE_EDGE, 'w') as gpio25edge:
#         gpio25edge.write('none')

#     # remove the pin from userspace
#     with open(GPIO_UNEXPORT_FILE, 'w') as unexport_file:
#         unexport_file.write(str(GPIO_INTERRUPT_PIN))

#     # disable the interrupt
#     int_enable_port = GPINTENA if port == GPIOA else GPINTENB
#     for board_index in range(MAX_BOARDS):
#         write(0, int_enable_port, board_index)



import sys
import errno
import subprocess
from distutils.core import setup


VERSION_FILE = 'pifacecommon/version.py'
PY3 = sys.version_info[0] >= 3


def get_version():
    if PY3:
        version_vars = {}
        with open(VERSION_FILE) as f:
            code = compile(f.read(), VERSION_FILE, 'exec')
            exec(code, None, version_vars)
        return version_vars['__version__']
    else:
        execfile(VERSION_FILE)
        return __version__


setup(
    name='pifacecommon',
    version=get_version(),
    description='The PiFace common functions module.',
    author='Thomas Preston',
    author_email='thomas.preston@openlx.org.uk',
    license='GPLv3+',
    url='https://github.com/piface/pifacecommon',
    packages=['pifacecommon'],
    long_description=open('README.md').read() + open('CHANGELOG').read(),
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3 or "
        "later (AGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='piface raspberrypi openlx',
)
